import json
import logging

log = logging.getLogger(__name__)


class Eventler(object):
    """listen for events on zmq in socket"""
    def __init__(self, in_socket, **kwargs):
        self.in_socket = in_socket

        # set additional arguments
        for name, value in kwargs.items():
            setattr(self, name, value)

    def handle_events(self):
        while True:
            kind, data = self.in_socket.recv_string().split(' ', 1)
            log.debug('got %s: %s' % (kind, data))
            data = json.loads(data)

            self.handle_event(kind, data)



"""Run HTTP REST API with live game data

Usage:
  munerator [options] restapi

Options:
  -v --verbose            Verbose logging
  -d --debug              Run in debug mode (autoreload)
  --http-listen ip:port   Host and port to have the HTTP server listen on [default: 0.0.0.0:8081]
  --database ip:port      Host and port for mongo database [default: 127.0.0.1:27017]

"""
import logging
import os

from docopt import docopt
from eve import Eve
from eve_mongoengine import EveMongoengine
from flask import current_app
from munerator.common.models import Games, Players, Votes, Gamemaps, PlaylistItems

log = logging.getLogger(__name__)


def root():
    return current_app.send_static_file('index.html')


def main(argv):
    args = docopt(__doc__, argv=argv)

    host, port = args['--database'].split(':')

    # app settings
    my_settings = {
        'MONGO_HOST': host,
        'MONGO_PORT': int(port),
        'MONGO_DBNAME': 'munerator',
        'DOMAIN': {'eve-mongoengine': {}},
        'X_DOMAINS': '*',
        'X_HEADERS': '*',
        'API_VERSION': 1,
        'URL_PREFIX': 'api'
    }

    # setup flask app
    static_folder = os.path.join(os.path.dirname(__file__), 'static')
    app = Eve(static_folder=static_folder, static_url_path='', settings=my_settings)

    # setup eve mongodb database
    ext = EveMongoengine(app)

    # add models
    ext.add_model([Players, Games, Votes, Gamemaps, PlaylistItems], resource_methods=['GET'])

    # register other urls
    app.route('/')(root)

    # run the app
    host, port = args['--http-listen'].split(':')
    app.run(host=host, port=int(port), debug=args['--debug'])



from eve import Eve
from eve_mongoengine import EveMongoengine
from .models import Players, Games, Votes, Gamemaps, PlaylistItems


def setup_eve_mongoengine(host='localhost', port=27017):
    # app settings
    my_settings = {
        'MONGO_HOST': host,
        'MONGO_PORT': int(port),
        'MONGO_PASSWORD': None,
        'MONGO_USERNAME': None,
        'MONGO_DBNAME': 'munerator',
        'DOMAIN': {'eve-mongoengine': {}},
    }
    app = Eve(settings=my_settings)

    # setup eve mongodb database
    ext = EveMongoengine(app)

    # add models
    ext.add_model([Players, Games, Votes, Gamemaps, PlaylistItems])



"""Change aspects about the current game

Usage:
  munerator [options] changer

Options:
  -v --verbose          Verbose logging
  --context-socket url  ZMQ socket for context events [default: tcp://quake.brensen.com:9002]
  --rcon-socket url     ZMQ socket for rcon commands [default: tcp://127.0.0.1:9005]

"""
import logging
from functools import partial

import zmq
from docopt import docopt
from munerator.common.eventler import Eventler
from munerator.games import Game

log = logging.getLogger(__name__)


class Changer(Eventler):

    def rcon(self, cmd):
        self.rcon_socket.send_string(cmd)

    def handle_event(self, kind, data):
        """
        Apply changes based on gameplay event. Eg. player joins -> increase fraglimit
        """

        if kind in ['clientbegin', 'clientdisconnect']:
            fraglimit = data.get('game_info', {}).get('fraglimit')
            num_players = data.get('game_info', {}).get('num_players')

            if all([fraglimit, num_players]):
                new_fraglimit = Game.get_fraglimit(num_players)
                log.debug('fraglimit:%s new_fraglimit:%s' % (fraglimit, new_fraglimit))

                if new_fraglimit > fraglimit:
                    log.info('new fraglimit %s' % new_fraglimit)
                    self.rcon('set fraglimit %s' % new_fraglimit)


def main(argv):
    args = docopt(__doc__, argv=argv)

    context = zmq.Context()

    in_socket = context.socket(zmq.SUB)
    in_socket.connect(args['--context-socket'])

    filters = ['clientbegin', 'clientdisconnect']
    add_filter = partial(in_socket.setsockopt, zmq.SUBSCRIBE)
    map(add_filter, filters)

    rcon_socket = context.socket(zmq.PUSH)
    rcon_socket.connect(args['--rcon-socket'])

    c = Changer(in_socket, rcon_socket=rcon_socket)
    c.handle_events()



"""
Reduce map into a playlist of most populair games to be played based current player votes, num. players, etc.
"""


def populate_playlist():
    pass

# votes mapreduce
# db.votes.mapReduce(
#     function(){
#         emit(this.mapname, this.vote)
#     },
#     function(key, values){
#         return 1,Array.sum(values)
#     },
#     {
#         query:{
#             gamemap:{
#                 $exists:true
#             }
#         },
#         out:{
#             inline:1
#         }
#     }
# )



"""Listen and output game event

Usage:
  munerator [options] listen [<event>...]

Options:
  -v --verbose          Verbose logging
  --context-socket url  ZMQ socket for context events [default: tcp://quake.brensen.com:9002]

"""
import json
import logging

import zmq
from docopt import docopt
from functools import partial


log = logging.getLogger(__name__)

exclude_keys = ['kind', 'timestamp']


def listen(in_socket, show_context=False):
    log.info('listening for game events')
    while True:
        msg = in_socket.recv_string()

        log.debug('got: %s' % msg)
        data = json.loads(msg.split(' ', 1)[-1])
        kind = data.get('kind')

        print(kind, ' '.join(['%s:%s' % (k, v) for k, v in data.items() if k not in exclude_keys]))


def main(argv):
    args = docopt(__doc__, argv=argv)

    context = zmq.Context()
    in_socket = context.socket(zmq.SUB)
    in_socket.connect(args['--context-socket'])

    if args['<event>']:
        log.debug('filters: %s' % ', '.join(args['<event>']))
        add_filter = partial(in_socket.setsockopt, zmq.SUBSCRIBE)
        map(add_filter, args['<event>'])
    else:
        in_socket.setsockopt(zmq.SUBSCRIBE, '')

    listen(in_socket)



#!/usr/bin/env python

"""
Python Quake 3 Library
$Id: pyquake3.py 456 2009-11-11 00:22:58Z prof $

A module to talk to Quake 3 servers from Python scripts.
Note that the current version was tested only with Urban
Terror servers, so there may be subtle problems when you
use it for actual Quake 3 or Open Arena servers. Also we
are not 100% backwards compatible to earlier releases of
pyquake3. Bummer! :-/ Most "missing" features are now in
the Connection class if you care to look.

http://misc.slowchop.com/misc/wiki/pyquake3
Copyright (C) 2006-2007 Gerald Kaszuba
http://www.urbanban.com/pyquake3/
Copyright (C) 2009 |ALPHA| Mad Professor <alpha.mad.professor@gmail.com>

Released under the GPL 2, see file COPYING for details.
"""

import re as RE
import socket as SO


class Error(Exception):

    """Base class for exceptions in this module."""
    pass


class ConnectionError(Error):

    """Error in network connection or protocol."""
    pass


class Connection(object):

    """
    Low level connection to a Quake 3 server. Note that we
    bridge two levels of abstraction here, networking and
    Quake 3 packet format. But who cares? :-D

    The trickiest part here is managing responses from the
    server as some commands generate multiple UDP packets!
    Check out receive_all() below for details.
    """

    PREFIX_LENGTH = 4
    PACKET_PREFIX = "\xff" * PREFIX_LENGTH

    def __init__(self, host, port, size=8192, timeout=1.0, retries=5):
        """
        Create a pseudo-connection to "host" and "port"; we
        try to give UDP communication a semblance of sanity.

        The internal UDP packet buffer will be "size" bytes,
        we'll wait "timeout" seconds for each response, and
        we'll retry commands "retries" times before failing.
        """
        # we neither want to deal with blocking nor with
        # timeouts that are plain silly in 2009...
        assert 0.1 <= timeout <= 4.0
        assert 4096 <= size <= 65536
        assert 1 <= retries <= 10
        self.socket = SO.socket(SO.AF_INET, SO.SOCK_DGRAM)
        # for SOCK_DGRAM connect() slips a default address
        # into each datagram; furthermore only data from the
        # "connected" address is delivered back; pretty neat
        self.socket.connect((host, port))
        self.socket.settimeout(timeout)
        self.host = host
        self.port = port
        self.size = size
        self.timeout = timeout
        self.retries = retries

    def send(self, data):
        """
        Send given data as a properly formatted packet.
        """
        self.socket.send("%s%s\n" % (Connection.PACKET_PREFIX, data))

    def receive(self):
        """
        Receive a properly formatted packet and return the
        unpacked (type, data) response pair. Note that one
        packet will be read, not multiple; use receive_all
        to get all packets up to a timeout.
        """
        packet = self.socket.recv(self.size)

        if packet.find(Connection.PACKET_PREFIX) != 0:
            raise ConnectionError("Malformed packet")

        first_line_length = packet.find("\n")
        if first_line_length == -1:
            raise ConnectionError("Malformed packet")

        response_type = packet[Connection.PREFIX_LENGTH:first_line_length]
        response_data = packet[first_line_length + 1:]

        return (response_type, response_data)

    def receive_all(self):
        """
        Receive a sequence of packets until a timeout
        exception. Check that all packets share a type,
        if so merge the data from all packets. Return
        the merged (type, data) response pair.
        """
        packets = []

        try:
            while True:
                packet = self.receive()
                packets.append(packet)
        except SO.timeout:
            # we timed out, so we'll assume that the
            # sequence of packets has ended; not sure
            # if this is a good idea...
            pass

        assert len(packets) > 0
        status, data = packets[0]
        for packet in packets[1:]:
            assert status == packet[0]
            data += packet[1]

        return (status, data)

    def command(self, cmd):
        """
        Execute given command and return (type, data)
        response pair. Commands will be retried for a
        number of times. (All response packets will be
        read and merged using receive_all.)
        """
        retries = self.retries
        response = None
        while retries > 0:
            self.send(cmd)
            try:
                response = self.receive_all()
            except Exception:
                # TODO: really catch Exception here? no
                # SO.error or something?
                retries -= 1
            else:
                return response
        raise ConnectionError("No response after %d attempts." % self.retries)

    def close(self):
        """Close connection."""
        self.socket.close()


class REs(object):

    """
    A container for regular expressions used to parse the
    result of certain well-known server commands. In best
    Perl tradition, they are totally unreadable. 8-O
    """
    # parse a player line from "getstatus" command
    # 11 50 "|ALPHA|MarvinTheSpud"
    GETSTATUS = RE.compile(r'^(-?)(\d+) (\d+) "(.*)"')
    # parse a player line from "rcon status" command
    # 2 0 70 |ALPHA| Mad Professor^7 0 127.0.0.1:35107 229 25000
    RCON_STATUS = RE.compile(
        r'\s*(\d+)\s+(-?)(\d+)\s+(\d+)\s+(.*)\^7\s+(\d+)\s+(\S*)\s+(\d+)\s+(\d+)')


class Player(object):

    """Record collecting information about a player."""

    def __init__(self):
        """Create empty record with lots of None fields."""
        # information from getstatus request
        self.frags = None
        self.ping = None
        self.name = None
        # information from rcon status request
        self.address = None
        self.slot = None
        self.lastmsg = None
        self.qport = None
        self.rate = None
        # information from dumpuser request
        self.guid = None
        self.variables = None

    def __str__(self):
        """Short summary of name, address, and guid."""
        return ("Player<name: %s; address: %s; guid: %s>" %
                (self.name, self.address, self.guid))


class Server(object):

    """Record collecting information about a server."""

    def __init__(self, filter_colors=True):
        """Create empty record with lots of None fields."""
        # meta information before connect
        self.filter = filter_colors
        self.host = None
        self.port = None
        # shortcuts to well-known variables
        self.name = None
        self.game = None
        self.map = None
        self.protocol = None
        self.version = None
        # dict of *all* server variables
        self.variables = {}
        # list of players
        self.players = []

    def address(self):
        """Helper to get "ip:port" for a server."""
        return "%s:%s" % (self.host, self.port)

    def get_address(self):
        """Compatibiltiy alias for address()."""
        return self.address()

    def command(self, command):
        """Wrapper calling Connection.command() for a server."""
        return self.connection.command(command)

    def filter_name(self, name):
        """Helper to remove Quake 3 color codes from player names."""
        result = ""
        i = 0
        while i < len(name):
            if name[i] == "^":
                i += 2
            else:
                result += name[i]
                i += 1
        return result

    def __str__(self):
        """Short summary of name, address, and map."""
        return ("Server<name: %s; address: %s; map: %s>" %
                (self.name, self.address(), self.map))


class Parser(object):

    """
    Mixin class to parse various server responses into
    useful information. Should be applied to subclasses
    of Server.
    """

    def parse_getstatus_variables(self, data):
        """
        Parse variables portion of getstatus response.
        The format is "\\key\\value\\key\\value..." and
        we turn that into a dictionary; selected values
        are also made fields.
        """
        data = data.split("\\")[1:]
        assert len(data) % 2 == 0
        keys = data[0::2]
        values = data[1::2]
        self.variables = dict(zip(keys, values))

        self.name = self.variables["sv_hostname"]
        self.game = self.variables["gamename"]
        self.map = self.variables["mapname"]
        self.protocol = self.variables["protocol"]
        self.version = self.variables["version"]

    def parse_getstatus_players(self, data):
        """
        Parse players portion of getstatus response.
        TODO
        """
        assert len(data) > 0
        self.players = []

        for record in data:

            match = REs.GETSTATUS.match(record)
            if match:
                negative, frags, ping, name = match.groups()
                if negative == "-":
                    frags = "-" + frags
                if self.filter:
                    name = self.filter_name(name)

                player = Player()
                player.frags = int(frags)
                player.ping = int(ping)
                player.name = name
                self.players.append(player)

    def parse_getstatus(self, data):
        """
        Parse server response to getstatus command. The
        first line of the response has lots of variables
        while the following lines have players.
        """
        data = data.strip().split("\n")

        variables = data[0].strip()
        players = data[1:]

        self.parse_getstatus_variables(variables)

        if len(players) > 0:
            self.parse_getstatus_players(players)

    def getstatus(self):
        """
        Basic server query for public information only.
        """
        status, data = self.connection.command("getstatus")
        if status == "statusResponse":
            self.parse_getstatus(data)

    def update(self):
        """
        Compatibiltiy alias for getstatus().
        """
        self.getstatus()

    def parse_rcon_status_players(self, data):
        """
        Parse players portion of RCON status response.
        TODO
        """
        assert len(data) > 0
        self.players = []

        for record in data:

            match = REs.RCON_STATUS.match(record)
            if match:
                slot, negative, frags, ping, name, lastmsg, address, qport, rate = match.groups(
                )
                if negative == "-":
                    frags = "-" + frags
                if self.filter:
                    name = self.filter_name(name)

                player = Player()
                player.slot = int(slot)
                player.frags = int(frags)
                player.ping = int(ping)
                player.name = name
                player.lastmsg = int(lastmsg)
                player.address = address
                player.qport = int(qport)
                player.rate = int(rate)
                self.players.append(player)

    def parse_rcon_status(self, data):
        """
        Parse RCON status response. There are at least
        three lines, the first is "map: bla" so we can
        get an updated map variable. The next two are
        the table header, all remaining ones (if any)
        are players, one player on each line.
        """
        data = data.strip().split("\n")
        mapname = data[0].strip().split(": ")[1].strip()

        self.variables["mapname"] = mapname
        self.map = mapname

        players = data[3:]
        if len(players) > 0:
            self.parse_rcon_status_players(players)

    def rcon_status(self):
        """
        TODO
        """
        status, data = self.rcon_command("status")
        if status == "print" and data.startswith("map"):
            self.parse_rcon_status(data)

    def rcon_update(self):
        """
        Compatibiltiy alias for rcon_status().
        """
        self.rcon_status()

    def parse_dumpuser(self, player, data):
        """
        Two header lines followed by "key value" lines
        separated by (lots of) spaces; spaces in values
        are present too, so we split at most once.
        TODO
        """
        data = data.strip().split("\n")[2:]
        variables = {}
        for record in data:
            # we split at most once to not lose spaces
            # inside a value (a name for example)
            separated = record.strip().split(None, 1)
            key = separated[0].strip()
            value = separated[1].strip()
            variables[key] = value

        # we need to avoid updating one player with
        # information for another, so we check for
        # some equalities before we believe the new
        # data to apply
        if player.address == variables["ip"] and player.rate == int(variables["rate"]):
            # alright, update the player object with new information
            player.variables = variables
            player.guid = variables["cl_guid"]

    def rcon_dumpuser_all(self):
        """
        TODO
        """
        for player in self.players:
            status, data = self.rcon_command("dumpuser %d" % player.slot)
            assert status == "print" and data.startswith("userinfo")
            self.parse_dumpuser(player, data)


class Guest(Server, Parser):

    """
    Server implementation that cannot perform any RCON
    commands. The right class if you are browsing some
    random servers.
    """

    def __init__(self, host, port, filter_colors=True):
        """
        TODO
        """
        Server.__init__(self, filter_colors)
        self.connection = Connection(host, port)
        self.host = host
        self.port = port


class Administrator(Server, Parser):

    """
    Server implementation that can perform any command
    an administrator can. The right class if you're in
    the business of writing admin interfaces.
    """

    def __init__(self, host, port, rcon_password, filter_colors=True):
        """
        TODO
        """
        Server.__init__(self, filter_colors)
        self.connection = Connection(host, port)
        self.host = host
        self.port = port
        self.rcon_password = rcon_password

    def rcon_command(self, command):
        """
        Execute an RCON command through the underlying
        connection and return the (type, data) response
        pair.
        """
        command = "rcon \"%s\" %s" % (self.rcon_password, command)
        status, data = self.connection.command(command)
        # TODO: why make this into an exception? the regular
        # command() method doesn't raise?
        if status.startswith(("Bad rcon", "No rcon")):
            raise ConnectionError(status.strip())
        return (status, data)


def PyQuake3(server, rcon_password=None, filter_colors=True):
    """
    Factory method for some backwards compatibility.
    """
    host, port = server.split(":")
    port = int(port)
    if rcon_password is None:
        return Guest(host, port, filter_colors)
    else:
        return Administrator(host, port, rcon_password, filter_colors)


def test_connection():
    c = Connection("tx.urbanban.com", 27960)

    status = c.command("getstatus")
    assert len(status) > 0
    print(status)

    status = c.command("rcon status")
    assert status[1].startswith("Bad rcon")
    print(status)

    c.close()

    try:
        print(c.command("getstatus"))
    except SO.error as e:
        assert e is not None
        print(e)

    try:
        d = Connection("tx.urbanban.com", 27969)
        d.command("getstatus")
    except ConnectionError as e:
        assert e is not None
        print(e)


def test_updates_and_players():
    # put your own server/password here to test
    a = Administrator("tx.urbanban.com", 27960, "you'll never guess :-D")
    a.update()
    for p in a.players:
        print(p)
    a.rcon_update()
    for p in a.players:
        print(p)
    a.rcon_dumpuser_all()
    for p in a.players:
        print(p)

if __name__ == '__main__':
    test_connection()
    test_updates_and_players()



"""Translate events

Usage:
  munerator [options] trans

Options:
  -v --verbose         Verbose logging
  --raw-socket url     ZMQ socket for raw logline [default: tcp://127.0.0.1:9000]
  --events-socket url  ZMQ socket for raw events [default: tcp://127.0.0.1:9001]

"""
from docopt import docopt
import zmq
import logging
log = logging.getLogger(__name__)

import re


translaters = [
    (r'[0-9: ]*InitGame: .*fraglimit\\(?P<fraglimit>[\w]+).*\\g_gametype\\(?P<gametype>[^\\])+\\mapname'
        r'\\(?P<mapname>[\w]+).*\\g_timestamp\\(?P<timestamp>[^\\]+)', 'initgame'),
    (r'[0-9: ]*ShutdownGame:.*', 'shutdowngame'),
    (r'[0-9: ]*say: (?P<player_name>[^:]+): (?P<text>.+)', 'say'),
    (r'[0-9: ]*ClientUserinfoChanged: (?P<client_id>\d+) n\\(?P<player_name>[^\\]+)\\t\\(?P<team_id>\d+).*id\\'
        r'(?P<guid>[\w]+)', 'clientuserinfochanged'),
    (r'.*client:(?P<client_id>\d+) health:(?P<health>[\d-]+).*', 'hit'),
    (r'[0-9: ]*Kill: [^:]+: (?P<killer>.+) killed (?P<killed>.+) by (?P<mod>[\w]+)', 'kill'),
    (r'[0-9: ]*Kill: [^:]+: (?P<killer_name>.+) killed (?P<player_name>.+) by (?P<mod>[\w]+)', 'killed'),
    (r'[0-9: ]*Kill: [^:]+: (?P<player_name>.+) killed (?P<killed_name>.+) by (?P<mod>[\w]+)', 'killer'),
    (r'[0-9: ]*ClientDisconnect: (?P<client_id>\d+)', 'clientdisconnect'),
    (r'[0-9: ]*ClientConnect: (?P<client_id>\d+)', 'clientconnect'),
    (r'[0-9: ]*ClientBegin: (?P<client_id>\d+)', 'clientbegin'),
    (r'[0-9: ]*PlayerScore: (?P<client_id>\d+) (?P<score>[\d\-]+):.*', 'playerscore'),
    (r'[0-9: ]*Item: (?P<client_id>\d+) item_quad.*', 'quad'),
    (r'[0-9: ]*ClientUserinfoChanged: (?P<guid>(?P<client_id>\d+)) n\\(?P<player_name>[^\\]+)\\t\\(?P<team_id>\d+).*'
        r'\\skill\\ (?P<skill>[\d\.]+).*id\\', 'clientuserinfochanged'),
    (r'cmd:getstatus response_type:statusResponse response:.*\\g_gametype\\(?P<gametype>[^\\])+\\mapname'
        r'\\(?P<mapname>[\w]+).*\\g_timestamp\\(?P<timestamp>[^\\]+)', 'getstatus'),
    (r"cmd:status response_type:print response:\s*(?P<client_id>[0-9]+)\s*(?P<score>[0-9]+)\s*(?P<ping>[0-9]+)\s*"
     r"(?P<player_name>.+)\^7\s*[^\s]+\s*(?P<address>[^\s]+)\s*[^\s]+\s*[^\s]+", 'clientstatus'),
    (r'cmd:dumpuser (?P<client_id>\d+) response_type:print response:cl_guid\s*(?P<guid>[\w]+)', 'dumpuser'),
    (r'cmd:dumpuser (?P<client_id>\d+) response_type:print response:headmodel\s*(?P<headmodel>[\w]+)', 'dumpuser'),
]

extras = {
    'getstatus': re.compile(r"\\([a-zA-Z0-9_]+)\\([^\\]+)"),
    'initgame': re.compile(r"\\([a-zA-Z0-9_]+)\\([^\\]+)"),
}

regexes = [(re.compile(r), k) for r, k in translaters]


def translate(line, regexes):
    for regex, kind in regexes:
        m = regex.match(line)
        # if match found
        if m:
            # get dictionary of capturegroups
            data = m.groupdict()

            # see if we need to get extra information out of the line
            if kind in extras:
                found = re.findall(extras.get(kind), line)
                data['extras'] = dict(found)
            yield kind, data


def handle_line(ts, line, out_socket):
    handled = False
    for kind, data in translate(line, regexes):
        data['kind'] = kind
        log.debug('out: %s' % data)
        handled = True

    if not handled:
        data = {
            'kind': 'unhandled',
            'raw': line
        }

    out_socket.send_json(data)


def eventstream(in_socket, out_socket):
    handled = 0
    while True:
        msg = in_socket.recv_string()
        ts, line = msg.split(' ', 1)

        log.debug('translating: ' + line)

        handle_line(ts, line, out_socket)

        handled += 1
        if not handled % 100:
            log.debug('handled another 100 messages (total %s)' % handled)


def main(argv):
    args = docopt(__doc__, argv=argv)

    context = zmq.Context()
    in_socket = context.socket(zmq.PULL)
    in_socket.bind(args['--raw-socket'])

    out_socket = context.socket(zmq.PUSH)
    out_socket.connect(args['--events-socket'])

    eventstream(in_socket, out_socket)



"""Run periodic tasks like finding assets (levelshot, images, etc), cleaning database.

Usage:
  munerator [options] periodic

Options:
  -v --verbose          Verbose logging
  -f --force            Force refresh of all assets
  --database ip:port    Host and port for mongo database [default: 127.0.0.1:27017]

"""
import logging
import urllib2
import hashlib

from docopt import docopt
from munerator.common.database import setup_eve_mongoengine
from munerator.common.models import Gamemaps, Games
from mongoengine.queryset import Q

log = logging.getLogger(__name__)


sources = {
    'levelshot': [
        'http://ws.q3df.org/images/levelshots/128x96/{mapname}.jpg',
        'http://gsdata.furver.se/thumbnails_quake3/{mapname}.jpg',
        'http://openarena.ws/svn/levelshots/{mapname}.jpg',
        'http://oa-community.com/map-manager/all/thumbnails/{mapname}.jpg',
    ],
    'images': [
        'http://ws.q3df.org/topview/{mapname}.jpg',
        'http://ws.q3df.org/images/levelshots/512x384/{mapname}.jpg',
        'http://gsdata.furver.se/levelshots_idtech3/{mapname}.jpg',
        'http://lvlworld.com/levels/{mapname}/{mapname}lg.jpg',
        'http://lvlworld.com/levels/map-{mapname}/map-{mapname}lg.jpg',
        'http://lvlworld.com/levels/{mapname}/{mapname}pan.jpg',

    ]
}

# hashes returned by server for not found images
notfoundhashes = [
    'dd2265645ae69a3c0f7b2f2340af18f0',
    '8d510c4aafcbf0238d56d84dfdad7a20',
    'b78ded5fb173185205e837ffa56c43e6',
]

default = 'http://ws.q3df.org/images/levelshots/512x384/'


def valid_asset(url):
    log.debug('url: %s' % url)
    try:
        response = urllib2.urlopen(url)
    except urllib2.HTTPError as e:
        response = e
        log.debug(response)

    if not response.code == 200:
        return False
    content = response.read()

    if not content:
        return False

    m = hashlib.md5()
    m.update(content)
    md5 = m.hexdigest()

    log.debug('hash: %s' % md5)

    if md5 in notfoundhashes:
        return False

    return True


def find_assets(force):
    if force:
        gamemaps = Gamemaps.objects()
    else:
        gamemaps = Gamemaps.objects(levelshot=None)

    for gamemap in gamemaps:
        log.debug('searching levelshot for %s' % gamemap.name)
        for source in sources['levelshot']:
            url = source.format(mapname=gamemap.name)
            if valid_asset(url):
                log.info('url %s, is valid, storing' % url)
                gamemap.update(set__levelshot=url)
                break
            else:
                log.debug('url %s is invalid' % url)
        else:
            # in case no levelshot is found
            gamemap.update(set__levelshot=default)

        for source in sources['images']:
            url = source.format(mapname=gamemap.name)
            if valid_asset(url):
                log.info('url %s, is valid, storing' % url)
                gamemap.update(add_to_set__images=url)
            else:
                log.debug('url %s is invalid' % url)


def clean_database():
    games = Games.objects(Q(current=False) | Q(current__exists=False), players__size=0)
    log.debug('removing games: %s' % games)
    games.delete()


def main(argv):
    args = docopt(__doc__, argv=argv)

    # setup database
    host, port = args['--database'].split(':')
    setup_eve_mongoengine(host, port)

    # run tasks
    find_assets(force=args['--force'])
    clean_database()



__all__ = ['Players', 'Games', 'Votes']

from mongoengine import (BooleanField, Document, IntField, DictField,
                         ListField, ReferenceField, StringField, URLField)


class Players(Document):
    guid = StringField()
    name = StringField()
    names = ListField(StringField())
    online = BooleanField()
    score = IntField()
    team = StringField()
    team_id = IntField()
    skill = StringField()
    address = StringField()
    bot = BooleanField()
    headmodel = StringField()

    # fields allowed to be updated by context info
    update_fields = ['name', 'online', 'score', 'team', 'team_id', 'skill', 'address', 'bot', 'headmodel']


class Gamemaps(Document):
    name = StringField()
    levelshot = URLField()
    images = ListField(URLField())
    gametypes = ListField(IntField())
    min_players = IntField(default=2)
    max_players = IntField(default=12)
    times_played = IntField(default=0)


class Votes(Document):
    game = ReferenceField('Games')
    player = ReferenceField('Players')
    gamemap = ReferenceField('Gamemaps')
    gametype = IntField()
    vote = IntField()


class Games(Document):
    timestamp = StringField()
    mapname = StringField()
    gamemap = ReferenceField('Gamemaps')
    gametype = IntField()
    players = ListField(ReferenceField('Players'))
    votes = ListField(ReferenceField('Votes'))
    state = StringField()
    start = StringField()
    stop = StringField()
    num_players = IntField()
    current = BooleanField()
    options = DictField()

    # fields allowed to be updated by context info
    update_fields = ['mapname', 'gametype', 'state', 'start', 'stop', 'num_players', 'current']


class PlaylistItems(Document):
    gamemap = ReferenceField('Gamemaps')
    votes = ListField(ReferenceField('Votes'))
    score = IntField(default=0)



from setuptools import setup, find_packages
import os


def read(fname):
    with open(fname) as fp:
        content = fp.read()
    return content

setup(
    name='munerator',
    version="0.12.5",
    description='Manager of OpenArena battles',
    long_description=read("README.rst"),
    author='Johan Bloemberg',
    author_email='mail@ijohan.nl',
    url='https://github.com/aequitas/munerator',
    install_requires=[
        'six',
        'docopt',
        'pyzmq',
        'webcolors',
        'python-Levenshtein',
        'python-dateutil',
        'rcfile'
    ],
    extras_require={
        'db': [
            'mongoengine',
            'eve_mongoengine==0.0.6-dev',
            'eve==0.4-dev',
            'flask',
        ]
    },
    dependency_links=[
        'https://github.com/hellerstanislav/eve-mongoengine/tarball/01e983b0dcf80b95cdbee799e78405559cc6cca7'
        '#egg=eve_mongoengine-0.0.6-dev',
        'https://github.com/aequitas/eve/tarball/25fc50e34ddc88ecd7ed14bd77124314f70622da#egg=eve-0.4-dev'
    ],
    packages=find_packages(),
    license=read("LICENSE"),
    include_package_data=True,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
    entry_points={
        'console_scripts': [
            "munerator = munerator:main",
        ]
    },
    data_files=[
        ('static', [os.path.join(dp, f) for dp, dn, filenames in os.walk('munerator/static') for f in filenames]),
    ],
)



"""Add game and player context to events

Usage:
  munerator [options] context

Options:
  -v --verbose          Verbose logging
  --events-socket url   ZMQ socket for raw events [default: tcp://127.0.0.1:9001]
  --context-socket url  ZMQ socket for context events [default: tcp://0.0.0.0:9002]
  --rcon-socket url     ZMQ socket for rcon commands [default: tcp://127.0.0.1:9005]

"""
from docopt import docopt
import zmq
import logging
import collections
import json
import time
import dateutil.parser

log = logging.getLogger(__name__)


team_id_map = ['', 'blue', 'red', 'spectator']
deduplicate = collections.deque(maxlen=5)


def get_dict_value_from_key_if_key_value(data, value_key, query_key, query_value):
    item = [v for k, v in data.items() if v.get(query_key) == query_value]
    if item:
        return item[0].get(value_key)


class GameContext(object):

    def __init__(self, in_socket, out_socket, rcon_socket):
        self.gameinfo = {}
        self.clients = {}

        self.in_socket = in_socket
        self.out_socket = out_socket
        self.rcon_socket = rcon_socket

    def eventstream(self):
        # prime commands for current game info
        self.rcon_socket.send_string('status')
        self.rcon_socket.send_string('getstatus')

        while True:
            data = self.in_socket.recv_json()
            log.debug('   in: %s' % data)

            if str(data) in deduplicate:
                log.debug('skip')
                continue
            deduplicate.append(str(data))

            contexted_data = self.handle_event(data)

            self.out_socket.send_string(
                "%s %s" % (contexted_data.get('kind'), json.dumps(contexted_data)))

    def handle_event(self, data):
        kind = data.get('kind')
        client_id = data.get('client_id')

        # for some events we need to translate player_name into client_id
        if not client_id and kind in ['say', 'killer', 'killed']:
            client_id = get_dict_value_from_key_if_key_value(
                self.clients, 'client_id', 'name', data.get('player_name'))
            data['client_id'] = client_id

        if client_id and client_id not in self.clients:
            self.clients[client_id] = {}

        if kind in ['initgame', 'getstatus']:
            self.gameinfo = {
                'mapname': data.get('mapname'),
                'gametype': data.get('gametype'),
                'timestamp': data.get('timestamp'),
                'fraglimit': data.get('fraglimit'),
                'start': dateutil.parser.parse(data.get('timestamp')).strftime('%s'),
                'current': True
            }
            if kind == 'initgame':
                self.gameinfo.update({
                    'num_players': 0,
                    'stop': ''
                })
                self.clients = {}
        elif kind in ['clientuserinfochanged', 'clientstatus']:
            log.debug('setting client info: %s' % client_id)
            self.clients[client_id].update({
                'name': data.get('player_name'),
                'id': data.get('guid'),
                'guid': data.get('guid'),
                'client_id': client_id,
                'team_id': data.get('team_id'),
                'skill': data.get('skill'),
                'team': team_id_map[int(data.get('team_id', 0))],
                'score': 0,
                'online': True,
                'bot': False,
            })
            if data.get('skill') or data.get('address') == 'bot':
                self.clients[client_id]['bot'] = True
            self.gameinfo['num_players'] = len(self.clients)
        elif kind == 'dumpuser':
            self.clients[client_id].update(data)
        elif kind == 'playerscore':
            log.debug('setting client score: %s' % client_id)
            self.clients[client_id]['score'] = data.get('score')
        elif kind == 'clientdisconnect':
            if client_id in self.clients:
                self.clients[client_id]['online'] = False

        if kind in ['clientstatus', 'clientconnect']:
            # on clientstatus also request dumpuser
            self.rcon_socket.send_string('dumpuser %s' % client_id)

        data['game_info'] = self.gameinfo
        data['client_info'] = self.clients.get(client_id, {})

        if kind == 'clientdisconnect':
            try:
                del self.clients[client_id]
            except KeyError:
                pass
        elif kind == 'shutdowngame':
            # add stop time
            self.gameinfo['stop'] = str(int(time.time()))
            self.gameinfo['current'] = False

            # reset current context
            self.gameinfo = {}
            self.clients = {}

        log.debug(' out: %s' % data)

        return data


def main(argv):
    args = docopt(__doc__, argv=argv)

    context = zmq.Context()
    in_socket = context.socket(zmq.PULL)
    in_socket.bind(args['--events-socket'])

    out_socket = context.socket(zmq.PUB)
    out_socket.bind(args['--context-socket'])

    # setup rcon socket
    rcon_socket = context.socket(zmq.PUSH)
    rcon_socket.connect(args['--rcon-socket'])

    gc = GameContext(in_socket, out_socket, rcon_socket)

    gc.eventstream()



class Game(object):
    specials = {
        "instagib": "set g_instantgib 2"
    }

    defaults = [
        "set g_warmup 0",
        "set g_doWarmup 0",
        "set g_spawnprotect 2000",
        "set g_speed 320",
        "set g_gravity 800",
        "set g_knockback 1000",
        "set map_restart 0",
        "set g_instantgib 0",
        "set g_vampire 0",
        "set g_regen 0",
        "set g_rockets 0",
        "set g_catchup 0"
    ]

    team_commands = [
        "set g_teamAutoJoin 0",
        "set g_teamForceBalance 1"
    ]

    catchup = "g_catchup 5"

    def __init__(self, mapname, gametype, num_players=3, special=None):
        self.mapname = mapname
        self.gametype = gametype
        self.num_players = num_players
        self.special = special

    @staticmethod
    def get_fraglimit(num_players):
        # increate fraglimit by 8 for every two players joining
        fraglimit = int(num_players) / 2 * 8
        fraglimit = fraglimit = max(10, min(fraglimit, 30))

        return fraglimit

    def mapstring(self, nextmap=None):
        cmds = list(self.defaults)

        # set gametype specific options
        if self.gametype in [3, 4, 5, 6, 7, 8, 9]:
            cmds += self.team_commands
        else:
            cmds.append(self.catchup)

        if self.special:
            cmds.append(self.specials[self.special])

        if self.special:
            cmds.append("set g_motd %s" % self.special)
        else:
            cmds.append("set g_motd ''")

        cmds.append("set fraglimit %s" % self.get_fraglimit(self.num_players))
        cmds.append("set g_gametype %s" % self.gametype)

        if nextmap:
            cmds.append('set nextmap "%s"' % nextmap)

        cmds.append("map %s" % self.mapname)

        return ";".join(cmds)

    def nextmapstring(self, nextmap=None):
        return 'set nextmap "%s";' % self.mapstring(nextmap).replace('"', '\\"')
